package jdk.internal.org.objectweb.asm.commons;

import java.io.DataOutput;
import java.security.MessageDigest;
import java.io.IOException;
import java.util.Arrays;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;
import jdk.internal.org.objectweb.asm.FieldVisitor;
import jdk.internal.org.objectweb.asm.MethodVisitor;
import java.util.ArrayList;
import java.util.Collection;
import jdk.internal.org.objectweb.asm.ClassVisitor;

public class SerialVersionUIDAdder extends ClassVisitor
{
    private boolean computeSVUID;
    private boolean hasSVUID;
    private int access;
    private String name;
    private String[] interfaces;
    private Collection<SerialVersionUIDAdder.Item> svuidFields;
    private boolean hasStaticInitializer;
    private Collection<SerialVersionUIDAdder.Item> svuidConstructors;
    private Collection<SerialVersionUIDAdder.Item> svuidMethods;
    
    public SerialVersionUIDAdder(ClassVisitor classVisitor) {
        this(327680, classVisitor);
        if (this.getClass() != SerialVersionUIDAdder.class) {
            throw new IllegalStateException();
        }
        return;
    }
    
    protected SerialVersionUIDAdder(int api, ClassVisitor classVisitor) {
        super(api, classVisitor);
        this.svuidFields = new ArrayList();
        this.svuidConstructors = new ArrayList();
        this.svuidMethods = new ArrayList();
    }
    
    @Override
    public void visit(int version, int n, String s, String signature, String superName, String[] interfaces) {
        this.computeSVUID = ((n & 0x200) == 0x0);
        if (this.computeSVUID) {
            this.name = s;
            this.access = n;
            System.arraycopy(interfaces, 0, this.interfaces = new String[interfaces.length], 0, interfaces.length);
        }
        super.visit(version, n, s, signature, superName, interfaces);
    }
    
    @Override
    public MethodVisitor visitMethod(int access, String s, String descriptor, String signature, String[] exceptions) {
        if (this.computeSVUID) {
            if ("<clinit>".equals(s)) {
                this.hasStaticInitializer = true;
            }
            int n = access & 0xD3F;
            if ((access & 0x2) == 0x0) {
                if ("<init>".equals(s)) {
                    this.svuidConstructors.add(new Item(s, n, descriptor));
                }
                else if (!"<clinit>".equals(s)) {
                    this.svuidMethods.add(new Item(s, n, descriptor));
                }
            }
        }
        return super.visitMethod(access, s, descriptor, signature, exceptions);
    }
    
    @Override
    public FieldVisitor visitField(int access, String s, String descriptor, String signature, Object value) {
        if (this.computeSVUID) {
            if ("serialVersionUID".equals(s)) {
                this.computeSVUID = false;
                this.hasSVUID = true;
            }
            if ((access & 0x2) == 0x0 || (access & 0x88) == 0x0) {
                this.svuidFields.add(new Item(s, access & 0xDF, descriptor));
            }
        }
        return super.visitField(access, s, descriptor, signature, value);
    }
    
    @Override
    public void visitInnerClass(String s, String outerName, String innerName, int n) {
        if (this.name != null && this.name.equals(s)) {
            this.access = n;
        }
        super.visitInnerClass(s, outerName, innerName, n);
    }
    
    @Override
    public void visitEnd() {
        if (this.computeSVUID && !this.hasSVUID) {
            try {
                this.addSVUID(this.computeSVUID());
            }
            catch (Throwable cause) {
                throw new RuntimeException("Error while computing SVUID for " + this.name, cause);
            }
        }
        super.visitEnd();
    }
    
    public boolean hasSVUID() {
        return this.hasSVUID;
    }
    
    protected void addSVUID(long l) {
        FieldVisitor visitField = super.visitField(24, "serialVersionUID", "J", null, Long.valueOf(l));
        if (visitField != null) {
            visitField.visitEnd();
        }
    }
    
    protected long computeSVUID() throws IOException {
        DataOutputStream dataOutputStream = null;
        long n = 0L;
        try {
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            dataOutputStream = new DataOutputStream(out);
            dataOutputStream.writeUTF(this.name.replace('/', '.'));
            dataOutputStream.writeInt(this.access & 0x611);
            Arrays.sort(this.interfaces);
            for (int i = 0; i < this.interfaces.length; ++i) {
                dataOutputStream.writeUTF(this.interfaces[i].replace('/', '.'));
            }
            writeItems(this.svuidFields, dataOutputStream, false);
            if (this.hasStaticInitializer) {
                dataOutputStream.writeUTF("<clinit>");
                dataOutputStream.writeInt(8);
                dataOutputStream.writeUTF("()V");
            }
            writeItems(this.svuidConstructors, dataOutputStream, true);
            writeItems(this.svuidMethods, dataOutputStream, true);
            dataOutputStream.flush();
            byte[] computeSHAdigest = this.computeSHAdigest(out.toByteArray());
            for (int j = Math.min(computeSHAdigest.length, 8) - 1; j >= 0; --j) {
                n = (n << 8 | (long)(computeSHAdigest[j] & 0xFF));
            }
        }
        finally {
            if (dataOutputStream != null) {
                dataOutputStream.close();
            }
        }
        return n;
    }
    
    protected byte[] computeSHAdigest(byte[] input) {
        try {
            return MessageDigest.getInstance("SHA").digest(input);
        }
        catch (Exception ex) {
            throw new UnsupportedOperationException(ex.toString());
        }
    }
    
    private static void writeItems(Collection<SerialVersionUIDAdder.Item> collection, DataOutput dataOutput, boolean b) throws IOException {
        int size = collection.size();
        SerialVersionUIDAdder.Item[] a = (SerialVersionUIDAdder.Item[])collection.toArray(new SerialVersionUIDAdder.Item[size]);
        Arrays.sort(a);
        for (int i = 0; i < size; ++i) {
            dataOutput.writeUTF(a[i].name);
            dataOutput.writeInt(a[i].access);
            dataOutput.writeUTF(b ? a[i].desc.replace('/', '.') : a[i].desc);
        }
    }
    
    private static class Item implements Comparable<SerialVersionUIDAdder.Item>
    {
        final String name;
        final int access;
        final String desc;
        
        Item(String name, int access, String desc) {
            this.name = name;
            this.access = access;
            this.desc = desc;
        }
        
        @Override
        public int compareTo(SerialVersionUIDAdder.Item item) {
            int n = this.name.compareTo(item.name);
            if (n == 0) {
                n = this.desc.compareTo(item.desc);
            }
            return n;
        }
        
        @Override
        public boolean equals(Object o) {
            return o instanceof SerialVersionUIDAdder.Item && this.compareTo((SerialVersionUIDAdder.Item)o) == 0;
        }
        
        @Override
        public int hashCode() {
            return (this.name + this.desc).hashCode();
        }
    }
}
